# Copyright 2009 Antoine d'Otreppe de Bouvette
#
# This file is part of the Aspyct library. see http://www.aspyct.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

if sys.version_info[0] == 2:
    def getFunctionLine(func):
        return func.func_code.co_firstlineno
    
    _imps = __import__('core.common', globals(), locals(), ['Aspect'])
else:
    def getFunctionLine(func):
        return func.__code__.co_firstlineno
    
    _imps = __import__('core.common', globals(), locals(), ['Aspect'], 1)

Aspect = _imps.Aspect
del _imps

__mods__ = {}
__errors__ = None

def runTests():
    result = {}
    for mod in __mods__:
        result[mod] = __mods__[mod].run()
    
    return result

def textRun():
    results = runTests()
    
    if not results:
        print("No test executed")
        return
    
    print("\n\nAspyct.unit test results\n")
    
    for mod in results:
        print("Module: %s"%mod)
        clean = True
        
        for func in results[mod]:
            if results[mod][func]:
                clean = False
                print("\tFunction %s"%func)
                for error in results[mod][func]:
                    print("\t\t%s @line %i"%(error[0], error[1]))
        
        if clean:
            print("\tClean")


def beforeTest(func):
    __getModule__(func.__module__).beforeTest = func
    return func

def afterTest(func):
    __getModule__(func.__module__).afterTest = func
    return func

def beforeMod(func):
    __getModule__(func.__module__).beforeMod = func
    return func

def afterMod(func):
    __getModule__(func.__module__).afterMod = func
    return func

def __report__(errStr, line=None):
    if line is None:
        line = sys._getframe(2).f_lineno
    
    __errors__.append((errStr, line))

def assertTrue(errStr, expr):
    if not expr:
        __report__(errStr)

def assertFalse(errStr, expr):
    if expr:
        __report__(errStr)

def assertEquals(errStr, expr1, expr2):
    if not expr1 == expr2:
        __report__(errStr)

def fail(errStr):
    __report__(errStr)

class Test(Aspect):
    def __init__(self, expected=None):
        Aspect.__init__(self)
        self.__expected = expected
        
    def __call__(self, func):
            test = Aspect.__call__(self, func)
            __getModule__(func.__module__).append(test)
            return test
    
    def atReturn(self, cd):
        if self.__expected:
            __report__('Error: expected ' + str(self.__expected)
                  + ' to be raised from ' + cd.function.__name__, getFunctionLine(cd.function))
    
    def atRaise(self, cd):
        if not self.__expected:
            __report__('Error: did not expect ' + str(cd.exception.__class__)
                  + ' to be raised from ' + cd.function.__name__, getFunctionLine(cd.function))
        elif not cd.exceptionIs(self.__expected):
            __report__('Error: expected ' + str(self.__expected)
                  + ' to be raised from ' + cd.function.__name__
                  + ', got ' + str(cd.exception.__class__), getFunctionLine(cd.function))
        
        cd.change()
        return None

class __TestModule(list):
    def __init__(self):
        list.__init__(self)
        self.__beforeMod = None
        self.__afterMod = None
        self.__beforeTest = None
        self.__afterTest = None
    
    def getBeforeMod(self):
        return self.__beforeMod
    def setBeforeMod(self, new):
        self.__beforeMod = new
    beforeMod = property(getBeforeMod, setBeforeMod)
    
    def getAfterMod(self):
        return self.__afterMod
    def setAfterMod(self, new):
        self.__afterMod = new
    afterMod = property(getAfterMod, setAfterMod)
    
    def getBeforeTest(self):
        return self.__beforeTest
    def setBeforeTest(self, new):
        self.__beforeTest = new
    beforeTest = property(getBeforeTest, setBeforeTest)
    
    def getAfterTest(self):
        return self.__afterTest
    def setAfterTest(self, new):
        self.__afterTest = new
    afterTest = property(getAfterTest, setAfterTest)
    
    def run(self):
        global __errors__
        result = {}
        
        if self.__beforeMod:
            self.__beforeMod()
        
        for test in self:
            __errors__ = []
            
            if self.__beforeTest:
                self.__beforeTest()
            
            test()
            
            if self.__afterTest:
                self.__afterTest()
            
            result[test.__name__] = __errors__
        
        if self.__afterMod:
            self.__afterMod()
        
        return result;
            
def __getModule__(modName):
    try:
        return __mods__[modName]
    except:
        new = __TestModule()
        __mods__[modName] = new
        return new

__all__ = ["runTests", "assertTrue", "assertFalse", "assertEquals", "fail", "Test",
           "beforeMod", "afterMod", "beforeTest", "afterTest", "textRun"]

