# Copyright 2009 Antoine d'Otreppe de Bouvette
#
# This file is part of the Aspyct library. see http://www.aspyct.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

if sys.version_info[0] == 2:
    import cPickle as pickle
    _imps = __import__('core.common', globals(), locals(), ['Aspect'])
else:
    import pickle
    _imps = __import__('core.common', globals(), locals(), ['Aspect'], 1)

Aspect = _imps.Aspect
del _imps

from threading import RLock

class Queryable(Aspect, list):
    ''' Apply this aspect to an __init__ function
    and created object will automatically be collected, enabling you to
    query them afterwards.
    
    If an error occurs during the object initialization, it will not be collected.
    This class inherits from list. Any action on list is doable on this.
    
    This aspect can search both attributes and properties
    
    Usage example:
        from Aspyct.utils import *
        
        q = Queryable()
        
        class SimpleClass:
            @q
            def __init__(self, a, b):
                self.__a = a
                self.__b = b
        
            def getA(self):
                return self.__a
        
            def getB(self):
                return self.__b
        
            a = property(getA, None, None, None)
        
            b = property(getB, None, None, None)
        
        SimpleClass(1, 1)
        SimpleClass(2, 4)
        SimpleClass(3, 2)
        SimpleClass(1, 3)
        SimpleClass(4, 2)
        SimpleClass(4, 2)
        
        print(q.query(a=4))
        
        print(q.query(a=1, b=3))
        
        q.save('test.dat')
        
        p = Queryable.open('test.dat')
        
        print(p.query(a=1, b=3))
    '''
    def atReturn(self, cd):
        self.append(cd.self)
    
    def query(self, **request):
        '''Use this function to query all the collected objects.
        
        Object.query(attr1="val1", attr2=19, ...)
        
        It will return you another Queryable object containing results.
        You may then query this new object again to refine your search.'''
        result = Queryable()
        
        for item in self:
            if Queryable.__scan(item, request):
                result.append(item)
        
        return result
    
    def save(self, path):
        '''Save the current Queryable and all of its entries to the specified path'''
        file = open(path, 'wb')
        pickle.dump(self, file)
        file.close()
    
    def clean(self, **request):
        '''Removes all entries corresponding to request from the list,
        or everything if request is empty'''
        for item in self:
            if Queryable.__scan(item, request):
                self.remove(item)
    
    @staticmethod
    def __scan(object, request):
        for attr in request:
            try:
                if getattr(object, attr) != request[attr]:
                    return False
            except:
                return False
        
        return True
    
    @staticmethod
    def open(path):
        '''This static method opens the given path,
        and returns the contained Queryable object'''
        file = open(path, "rb")
        return pickle.load(file)


class Logger(Aspect):
    ''''This is an abstract logging class. Subclass it and implement the "log" method
    See also TextLogger, StdoutLogger and FileLogger for generic implementations'''
    CALL = 1
    RAISE = 2
    RETURN = 3
    
    def __init__(self, logCall=True, logRaise=True, logReturn=True):
        Aspect.__init__(self)
        
        self.__apply= False
        if logCall:
            self.atCall = self.__atCall
            self.__apply= True
        if logRaise:
            self.atRaise = self.__atRaise
            self.__apply= True
        if logReturn:
            self.atReturn = self.__atReturn
            self.__apply= True
    
    def __call__(self, func):
        if self.__apply:
            return Aspect.__call__(self, func)
        else:
            return func
        
    def __atCall(self, cd):
        self.log(Logger.CALL, cd)
    
    def __atRaise(self, cd):
        self.log(Logger.RAISE, cd)
    
    def __atReturn(self, cd):
        self.log(Logger.RETURN, cd)


class TextLogger(Logger):
    '''This is an abstract class for text-based logging
    Subclass it, and implement your "write" method, receiving one argument.
    This argument contains the text to be logged'''
    def log(self, when, cd):
        
        if cd.cls:
            cls = cd.cls.__name__ + '(' + str(id(cd.self)) + ').'
        else:
            cls = ''
        
        if when == Logger.CALL:
            kwargs = []
            for item in cd.kwargs:
                kwargs.append(str(item) + '=' + str(cd.kwargs[item]))
            
            string = "%s%s(%s, %s)"% \
                (cls, cd.function.__name__, ", ".join(cd.args), ", ".join(kwargs))
        elif when == Logger.RAISE:
            exc = cd.exception
            string = "%s%s raise %s(%s)"% \
                (cls, cd.function.__name__, exc.__class__.__name__, str(exc))
        elif when == Logger.RETURN:
            string = "%s%s return %s"% \
                (cls, cd.function.__name__, cd.returned)
            
        self.write(string)


class StdoutLogger(TextLogger):
    '''Logs to stdout
    Usage example:
        log = StdoutLogger()
        
        @log
        def test():
            pass
        
        @log
        def test2(arg1, arg2):
            return arg2
    '''
    def write(self, text):
        print(text)


class FileLogger(TextLogger):
    '''Logs to the given file path
    Usage example:
        log = FileLogger("path/to/logfile")
        
        @log
        def test():
            pass
        
        @log
        def test2(arg1, arg2):
            return arg2
    '''
    def __init__(self, path, logCall=True, logRaise=True, logReturn=True):
        TextLogger.__init__(self, logCall, logRaise, logReturn)
        self.__file = open(path, 'a')
        self.__write = self.__file.write
        self.__flush = self.__file.flush
    
    def write(self, text):
        self.__write(text)
        self.__write('\n')
        self.__flush()
    
    def close(self):
        self.__file.close()

class Synchronized(Aspect):
    '''This aspect is similar to Java's "synchronized" statement.
    It uses RLock to lock an object, so that the same thread may
    acquire lock several times transparently.
    
    Usage examples:
    
    You can synchronized whole functions or methods
        >>> class NS:
        ...     @Synchronized() # With no argument, synchronizes on the "self" object
        ...     def test(self):
        ...         print "Hello"
        ... 
        >>> a = NS()
        >>> class MyThread(Thread):
        ...     @Synchronized(a) # Synchronizes on the "a" object
        ...     def run(self):
        ...         sleep(10)
        ... 
        >>> MyThread().start()
        >>> a.test()
        
    Or only small blocks
        >>> class MyThread2(Thread):
        ...     def run(self):
        ...         # Do some stuff
        ...         
        ...         # Create a synchronized block
        ...         @Synchronized(a)
        ...         def sync():
        ...             sleep(10)
        ...         sync()
        ...         
        ...         # Do other stuff
        ... 
        >>> MyThread2().start()
        >>> a.test()
    
    This class also provides "getLock" as a static method, for fine-grained control
        >>> class NS2: 
        ...     def test(self):
        ...         print "Hello"
        ... 
        >>> b = NS2()
        >>> class MyThread3(Thread):
        ...     def run(self):
        ...         l = Synchronized.getLock(b)      
        ...         l.acquire()
        ...         b.test()
        ...         l.release()
        ... 
        >>> Synchronized.getLock(b).acquire()
        True
        >>> MyThread3().start()
        >>> Synchronized.getLock(b).release()
        Hello

    '''
    def __init__(self, object=None):
        if object:
            self.__object = object
    
    def atCall(self, cd):
        try:
            obj = self.__object
        except AttributeError:
            self.__object = cd.self
            obj = cd.self
        
        try:
            obj.__lock__.acquire()
        except AttributeError:
            l = RLock()
            obj.__lock__ = l
            l.acquire()
    
    def atRaise(self, cd):
        self.__object.__lock__.release()
    
    def atReturn(self, cd):
        self.__object.__lock__.release()
    
    @staticmethod
    def getLock(object):
        try:
            return object.__lock__
        except AttributeError:
            l = RLock()
            object.__lock__ = l
            return l


__all__ = ["Queryable", "Logger", "TextLogger", "StdoutLogger",
           "FileLogger", "Synchronized"]
